#!/bin/env python

"""
Plot skew-T diagram of converted soundings
"""
import sys
import os.path
import argparse
import logging
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

import numpy as np
import metpy.calc as mpcalc
from metpy.cbook import get_test_data
from metpy.plots import add_metpy_logo, Hodograph, SkewT
from metpy.units import units
import xarray as xr


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--inputfile", metavar="INPUT_FILE",
                        help="Converted sounding file (netCDF)", default=None,
                        required=True)

    parser.add_argument("-o", "--outputfile", metavar="/some/example/path/filename.pdf",
                        help="Output filename for skewT diagram (all file endings"\
                             " of matplotlib are supported",
                        default=None,
                        required=False)

    parser.add_argument('-v', '--verbose', metavar="DEBUG",
                        help='Set the level of verbosity [DEBUG, INFO,'
                        ' WARNING, ERROR]',
                        required=False, default="INFO")

    parsed_args = vars(parser.parse_args())

    return parsed_args


def setup_logging(verbose):
    assert verbose in ["DEBUG", "INFO", "WARNING", "ERROR"]
    logging.basicConfig(
        level=logging.getLevelName(verbose),
        format="%(levelname)s - %(name)s - %(funcName)s - %(message)s",
        handlers=[
            logging.FileHandler("{}.log".format(__file__)),
            logging.StreamHandler()
        ])


def main():
    args = get_args()
    setup_logging(args['verbose'])

    # Define input file
    file = args['inputfile']
    output = args['outputfile']

    ds = xr.open_dataset(file)

    ds_sel = ds.isel({'sounding': 0})
    ds_sel = ds_sel.sortby(ds_sel.pressure, ascending=False)

    p = ds_sel.pressure.values
    T = ds_sel.temperature.values
    Td = ds_sel.dewPoint.values
    wind_speed = ds_sel.windSpeed.values
    wind_dir = ds_sel.windDirection.values

    # Filter nans
    idx = np.where((np.isnan(T)+np.isnan(Td)+np.isnan(p)+
                    np.isnan(wind_speed)+np.isnan(wind_dir)) == False, True, False)
    p = p[idx]
    T = T[idx]
    Td = Td[idx]
    wind_speed = wind_speed[idx]
    wind_dir = wind_dir[idx]

    # Add units
    p = p * units.hPa
    T = T * units.degC
    Td = Td * units.degC
    wind_speed = wind_speed * (units.meter/units.second)
    wind_dir = wind_dir * units.degrees

    u, v = mpcalc.wind_components(wind_speed, wind_dir)

    lcl_pressure, lcl_temperature = mpcalc.lcl(p[0], T[0], Td[0])

    parcel_prof = mpcalc.parcel_profile(p, T[0], Td[0]).to('degC')

    # Create a new figure. The dimensions here give a good aspect ratio
    fig = plt.figure(figsize=(9, 9))
    skew = SkewT(fig, rotation=30)

    # Plot the data using normal plotting functions, in this case using
    # log scaling in Y, as dictated by the typical meteorological plot
    skew.plot(p, T, 'r')
    skew.plot(p, Td, 'g')
    # Plot only specific barbs to increase visibility
    pressure_levels_barbs = np.logspace(0.1, 1, 50)*100

    def find_nearest(array, value):
        array = np.asarray(array)
        idx = (np.abs(array - value)).argmin()
        return array[idx]

    # Search for levels by providing pressures
    # (levels is the coordinate not pressure)
    pres_vals = ds_sel.pressure.values[idx]
    closest_pressure_levels = np.unique([find_nearest(pres_vals, p_) for p_ in pressure_levels_barbs])
    _, closest_pressure_levels_idx, _ = np.intersect1d(pres_vals, closest_pressure_levels, return_indices=True)

    p_barbs = ds_sel.pressure.isel({'levels': closest_pressure_levels_idx}).values * units.hPa
    wind_speed_barbs = ds_sel.windSpeed.isel({'levels': closest_pressure_levels_idx}).values * (units.meter/units.second)
    wind_dir_barbs = ds_sel.windDirection.isel({'levels': closest_pressure_levels_idx}).values * units.degrees
    u_barbs, v_barbs = mpcalc.wind_components(wind_speed_barbs, wind_dir_barbs)

    # Find nans in pressure
    # p_non_nan_idx = np.where(~np.isnan(pres_vals))
    skew.plot_barbs(p_barbs, u_barbs, v_barbs)

    skew.ax.set_ylim(1020, 100)
    skew.ax.set_xlim(-50, 40)

    # Plot LCL as black dot
    skew.plot(lcl_pressure, lcl_temperature, 'ko', markerfacecolor='black')

    # Plot the parcel profile as a black line
    skew.plot(pres_vals, parcel_prof, 'k', linewidth=2)

    # Shade areas of CAPE and CIN
    skew.shade_cin(pres_vals, T, parcel_prof)
    skew.shade_cape(pres_vals, T, parcel_prof)

    # Plot a zero degree isotherm
    skew.ax.axvline(0, color='c', linestyle='--', linewidth=2)

    # Add the relevant special lines
    skew.plot_dry_adiabats()
    skew.plot_moist_adiabats()
    skew.plot_mixing_lines()

    # Create a hodograph
    # Create an inset axes object that is 40% width and height of the
    # figure and put it in the upper right hand corner.
    ax_hod = inset_axes(skew.ax, '40%', '40%', loc=1)
    h = Hodograph(ax_hod, component_range=80.)
    h.add_grid(increment=20)
    h.plot_colormapped(u, v, wind_speed)  # Plot a line colored by wind speed

    # Set title
    sounding_name = ds_sel.sounding.values
    sounding_name_str = str(sounding_name.astype('str'))
    skew.ax.set_title('{sounding}'.format(
        sounding=sounding_name_str))

    if output is None:
        output = str(os.path.basename(file).split('.')[0])+'.pdf'
    logging.info('Write output to {}'.format(output))
    plt.savefig(output)


if __name__ == '__main__':
    main()



"""
Script to produce quicklooks of radiosonde type RS41, read from netcdf-files
converted using L1-rs41.py

Original version by: Sabrina Schnitt
"""

import glob
import sys
import getopt
import os
import re
import math
import logging
import netCDF4 as nc
from matplotlib.ticker import AutoMinorLocator
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.basemap import Basemap


def read_ncfile(ncfile):
    '''
    routine reads variables and attributes of ncfile generated by L1-rs41.py.
    INPUT: path+filename of netcdf-file
    OUTPUT: dictionnary with all variables and attributes.
    '''

    data = {}

    ncf = nc.Dataset(ncfile)

    for k in ncf.variables.keys():
        try:
            data[k] = ncf.variables[k][0, :]
        except ValueError:
            data[k] = ncf.variables[k][:]

    for a in ncf.ncattrs():
        data[a] = ncf.getncattr(a)

    specs = {}

    specs['location'] = ncf.location
    specs['tempres'] = ncf.resolution
    specs['date'] = ncf.date_YYYYMMDD
    specs['time'] = ncf.time_of_launch_HHmmss
    specs['type'] = ncf.instrument
    # Extract platform short name from the global attribute
    specs['platform_short'] = re.search(r"\(([A-Za-z0-9_]+)\)",
                                        ncf.platform_name).group(1)
    # Extract main flight direction from variable ascentRate
    most_common_vertical_movement = np.argmax(
        [np.count_nonzero(data['ascentRate'] > 0),
         np.count_nonzero(data['ascentRate'] < 0)])
    if most_common_vertical_movement == 0:
        # Mostly ascent rates
        specs['direction'] = 'AscentProfile'
    elif most_common_vertical_movement == 1:
        # Mostly descending rates
        specs['direction'] = 'DescentProfile'
    else:
        logging.warning('Main flight direction (ascent/descent) of instrument'
                        ' could not be identified!')
        specs['direction'] = 'Unknow'

    return data, specs


def plot_ptrh(data, specs, outputpath):
    '''
    routine plots vertical profiles of temperature, pressure, rel humidity and
    saves plot as .png
    INPUT:
        - data: dictionnary with data (eg filled by read_ncfile())
        - specs: dictionnary with filename specifications
            (filled by read_ncfile())
        - outputpath: path where png will be stored in.
    OUTPUT: .png file stored in outputpath
    '''
    logging.info('now plotting pressure, temperature, rel humidity sounding.........')

    # define outputname of .png-file:
    variable = 'ptrelh'
    outputname = '{platform}_{instrument}{direction}_{variable}_{date}_{tempres}.png' .format(
        platform=specs['platform_short'],
        instrument=specs['type'].replace(' ', '').replace('_', ''),
        direction=specs['direction'],
        variable=variable,
        date=specs['date']+'_'+specs['time'],
        tempres=specs['tempres'].replace(' ', ''))

    fig, ax = plt.subplots(1, 3, sharey=True, figsize=(8, 6))

    # plot temperature, pressure, humidity in three panels:
    ax[0].plot(data['temperature'], data['altitude'], '.-k', markersize=1)
    ax[1].plot(data['pressure'], data['altitude'], '.-k', markersize=1)
    ax[2].plot(data['humidity'], data['altitude'], '.-k', markersize=1)

    # do some cosmetics regarding the layout, axislabels, etc.:
    for i in range(3):
        # switch off some spines:
        ax[i].spines['top'].set_visible(False)
        ax[i].spines['right'].set_visible(False)
        ax[i].spines['left'].set_visible(False)
        ax[i].grid(axis='y', linestyle='-', color='gray')

        # set height axis to start at 0m:
        ax[i].set_ylim(0, ax[i].get_ylim()[-1])
        # major minor ticks:
        ax[i].xaxis.set_minor_locator(AutoMinorLocator())
        ax[i].yaxis.set_minor_locator(AutoMinorLocator())
        ax[i].xaxis.set_major_locator(plt.MaxNLocator(4))
        # switch off major ticks top and right axis:
        ax[i].tick_params(top=False, right=False)
        # switch off minor ticks for top axis:
        ax[i].tick_params(axis='x', which='minor', top=False)
        # make labels larger for all ticks:
        ax[i].tick_params(axis='both', labelsize=14)

    ax[0].spines['left'].set_visible(True)
    ax[0].tick_params(axis='y', right=False, which='minor')
    ax[1].tick_params(left=False)
    ax[1].tick_params(axis='y', right=False, left=False, which='minor')
    ax[2].tick_params(left=False)
    ax[2].spines['right'].set_visible(True)
    ax[2].yaxis.set_ticks_position('right')
    ax[2].yaxis.set_label_position('right')

    # set the relh panel always to values between 0 and 100:
    ax[2].set_xlim(0, 100)
    # and the pressure to max 1100 hPa:
    ax[1].set_xlim(ax[1].get_xlim()[0], 1100)

    # axis labels:
    ax[0].set_ylabel('Altitude [m]', fontsize=14)
    ax[2].set_ylabel('Altitude [m]', fontsize=14)

    ax[0].set_xlabel('Temperature [$^\circ$C]', fontsize=14)
    ax[1].set_xlabel('Pressure [hPa]', fontsize=14)
    ax[2].set_xlabel('Rel Humidity [%]', fontsize=14)

    plt.subplots_adjust(top=0.9, right=0.85, left=0.15)

    fig.suptitle('%s, %s %sUTC' % (specs['location'],
                                   specs['date'],
                                   data['time_of_launch_HHmmss'][:-2]),
                 fontsize=18)

    fig.savefig(outputpath+outputname)

    logging.info('{} profiles saved at {}'.format(variable,
                                                  outputpath+outputname))


def plot_wind(data, specs, outputpath):
    '''
    routine plots vertical profiles of wind speed and direction
    and saves plot as .png
    INPUT:
        - data: dictionnary with data (eg filled by read_ncfile())
        - specs: dictionnary with filename specifications (filled
            by read_ncfile())
        - outputpath: path where png will be stored in.
    OUTPUT: .png file stored in outputpath
    '''
    logging.info('now plotting wind speed and direction sounding.........')
    # define outputname of .png-file:
    variable = 'wind'
    outputname = '{platform}_{instrument}{direction}_{variable}_{date}_{tempres}.png' .format(
        platform=specs['platform_short'],
        instrument=specs['type'].replace(' ', '').replace('_', ''),
        direction=specs['direction'],
        variable=variable,
        date=specs['date']+'_'+specs['time'],
        tempres=specs['tempres'].replace(' ', ''))

    fig, ax = plt.subplots(1, 2, sharey=True, figsize=(8, 6))

    # plot the data into subpanels:
    ax[0].plot(data['windSpeed'], data['altitude'], '.-k', markersize=1)
    ax[1].plot(data['windDirection'], data['altitude'], '.-k', markersize=1)

    # general cosmetics:
    for i in range(2):
        ax[i].spines['top'].set_visible(False)
        ax[i].spines['right'].set_visible(False)
        ax[i].spines['left'].set_visible(False)
        ax[i].grid(axis='y', linestyle='-', color='gray')
        ax[i].set_ylim(0, ax[i].get_ylim()[-1])
        ax[i].xaxis.set_minor_locator(AutoMinorLocator())
        ax[i].yaxis.set_minor_locator(AutoMinorLocator())
        ax[i].xaxis.set_major_locator(plt.MaxNLocator(4))
        ax[i].tick_params(top=False, right=False)
        # axis labels:
        ax[i].set_ylabel('Altitude [m]', fontsize=14)
        # switch off minor ticks for top axis:
        ax[i].tick_params(axis='x', which='minor', top=False)
        # make labels larger for all ticks:
        ax[i].tick_params(axis='both', labelsize=14)

    # switch off some ticks and labels and spines manually.
    ax[0].spines['left'].set_visible(True)
    ax[1].tick_params(left=False)
    ax[0].tick_params(right=False, which='minor', axis='y')

    ax[1].spines['right'].set_visible(True)
    ax[1].yaxis.set_ticks_position('right')
    ax[1].yaxis.set_label_position('right')

    # set wind direction axis to valid range:
    ax[1].set_xlim(0, 360)

    ax[0].set_xlabel('Wind Speed [m s$^{-1}$]', fontsize=14)
    ax[1].set_xlabel('Wind Direction [$^\circ$]', fontsize=14)

    plt.subplots_adjust(top=0.9, right=0.85, left=0.15)
    fig.suptitle('%s, %s %sUTC' % (specs['location'],
                                   specs['date'],
                                   data['time_of_launch_HHmmss'][:-2]),
                 fontsize=18)

    fig.savefig(outputpath+outputname)

    logging.info('Wind profile saved at {}'.format(outputpath+outputname))


def plot_map(data, specs, outputpath):
    '''
    routine plots balloon flight on a map.
    INPUT:
        - data: dictionnary with data (eg filled by read_ncfile())
        - specs: dictionnary with filename specifications
            (filled by read_ncfile())
        - outputpath: path where png will be stored in.
    OUTPUT: .png file stored in outputpath.
    REQUIRES: basemap-data-hires package to be installed.
    '''
    logging.info('now plotting map of sounding.........')
    # define outputname of .png-file:
    variable = 'map'
    outputname = '{platform}_{instrument}{direction}_{variable}_{date}_{tempres}.png' .format(
        platform=specs['platform_short'],
        instrument=specs['type'].replace(' ', '').replace('_', ''),
        direction=specs['direction'],
        variable=variable,
        date=specs['date']+'_'+specs['time'],
        tempres=specs['tempres'].replace(' ', ''))

    fig = plt.figure(figsize=(8, 6))

    # determine the boundaries of the map from sounding lon and lat:
    maxlon = math.ceil(np.max(data['longitude'])/0.5)*0.5
    minlon = math.floor(np.min(data['longitude'])/0.5)*0.5

    maxlat = math.ceil(np.max(data['latitude'])/0.5)*0.5
    minlat = math.floor(np.min(data['latitude'])/0.5)*0.5

    # set up basemap projection
    try:
        m = Basemap(projection='cyl', resolution='h', llcrnrlat=minlat,
                    urcrnrlat=maxlat, llcrnrlon=minlon, urcrnrlon=maxlon,
                    area_thresh=1)
    except OSError:
        logging.warning('High resolution map data has not been installed and'
                        ' the low resolution resolution will be used. For the'
                        ' hight resolution install with e.g. conda install -c'
                        ' conda-forge basemap-data-hires')
        m = Basemap(projection='cyl', resolution='l', llcrnrlat=minlat,
                    urcrnrlat=maxlat, llcrnrlon=minlon, urcrnrlon=maxlon,
                    area_thresh=1)
    # plot a topography on top:
    m.etopo(alpha=0.4)

    # coastlines, countries, boundary, background-color, gridlines
    m.drawcoastlines()
    m.drawcountries()
    m.drawmapboundary()
    m.shadedrelief()
    m.fillcontinents(color='#00a500')
    m.drawparallels(np.arange(10, 70, 0.25), labels=[1, 1, 0, 0])
    m.drawmeridians(np.arange(-100, 0, 0.25), labels=[0, 0, 0, 1])

    # plot balloon path:
    x, y = m(data['longitude'], data['latitude'])
    m.plot(x, y, '-k')

    # plot launch position as red square:
    m.plot(float(data['longitude'][0]),
           float(data['latitude'][0]),
           'sr',
           markersize=5)

    # and the figure title:
    plt.title('%s, %s %sUTC' % (specs['location'], specs['date'],
                                data['time_of_launch_HHmmss'][:-2]))

    fig.savefig(outputpath+outputname)

    logging.info('Map saved at {}'.format(outputpath+outputname))


def setup_logging(verbose):
    assert verbose in ["DEBUG", "INFO", "WARNING", "ERROR"]
    logging.basicConfig(
        level=logging.getLevelName(verbose),
        format="%(levelname)s - %(name)s - %(funcName)s - %(message)s",
        handlers=[
            logging.FileHandler("{}.log".format(__file__)),
            logging.StreamHandler()
        ])


def main():
    # options for the skript: date (yymmddhh); inputfilename; outputpath;
    # inputpath
    # (needs to be specified if date is used);
    # defaults: outputpath: './'; inputpath: './'
    # either date+inputpath or inputfilename incl path need to be specified.
    # everything else is optional.

    inputfile = ''
    date = ''
    outputpath = './'  # default setting
    inputpath = './'  # default setting
    ncfile = None

    setup_logging('INFO')

    try:
        opts, args = getopt.getopt(sys.argv[1:],
                                   'd:n:o:i:h',
                                   ['date=', 'inputncfile=', 'outputpath=',
                                    'inputpath=', 'help'])
    except getopt.GetoptError:
        print('usage: python make_quicklooks_rs41.py -i <inputpath> -d <yymmddhh>'
              ' -n <inputncfile> -o <outputpath> ')
        sys.exit(2)

    for opt, arg in opts:
        if opt in ('-h', "--help"):  # help option
            print('usage: python make_quicklooks_rs41.py -i <inputpath> -d <yymmddhh> -n <inputncfile> -o <outputpath>')
            print('specify either complete input netcdf filename (-n) or date (-d) of sounding from which file be searched in -i inputpath. ')
            print('default inputpath: current directory. specify with -i option if different. if used together with -d, make sure -i is given first in call.')
            print('default outputpath: current directory. if -o is specified, outputpath is created if not yet existant.')
            sys.exit()

        elif opt in ("-i", "--inputpath"):
            inputpath = arg
        elif opt in ("-d", "--date"):
            try:
                ncfile = glob.glob(inputpath + '*%s*.nc' % arg)[0]
            except IndexError:
                logging.error('couldnt find your specified input: check date'
                              ' or/and inputpath selection.')
                sys.exit()
        elif opt in ("-n", "--inputncfile"):
            ncfile = arg
            if not os.path.isfile(ncfile):
                logging.error('couldnt find your specified inputfile.')
                sys.exit()

        elif opt in ("-o", "--outputpath"):
            outputpath = arg
            # check if there's a backslash after the outputpath-argument:
            if outputpath[-1] != '/':
                outputpath = outputpath+'/'
            if not os.path.isdir(outputpath):
                os.mkdir(outputpath)

        if ncfile is None:
            logging.error('Input file must be defined with either'
                          ' --inputncfile or --inputpath and --date')
            sys.exit()

    logging.info('plotting sounding file %s' % ncfile)

    # read netcdf-variables into dictionnary:
    radiosonde_data, radiosonde_specs = read_ncfile(ncfile)

    # make first quicklook: p, relh, T- profiles.
    plot_ptrh(radiosonde_data, radiosonde_specs, outputpath)

    # now also plot wind speed and direction:
    plot_wind(radiosonde_data, radiosonde_specs, outputpath)

    # also plot the sounding onto a map: REQUIRES BASEMAP-DATA-HIRES package
    # to be installed (e.g. through
    # conda install -c conda-forge basemap-data-hires)

    plot_map(radiosonde_data, radiosonde_specs, outputpath)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Thermodynamics functions

From Chris Bretherton, Bolton 1980, etc.

Simon de Szoeke
Created on Mon Aug 27 21:30:59 2018

@author: sdeszoek
"""

import numpy as np

#  Important global constants (Emanuel, Appendix 2)
g     = 9.8
stefan= 5.67e-8
r_earth = 6370*1000
Omega   = 2*np.pi/86400

Rd    = 287.04
Rv    = 461.50
Cp    = 1005.7
Cpv   = 1870
Cw    = 4190
L     = 2.501e6
delta = 0.608 # Rv/Rd - 1
CtoK  = 273.15

# Blackbody radiative constants
kb=1.381e-23     # J/K   Boltzmann
hplanck=6.6261e-34   # J s   Planck
clight=2.998e8   # m/s   speed of light
hc=1.98645e-25   # J m   Planck*speed of light

# Usually functions describing water and water vapor depend on T in Celsius
# and functions of generalized gases, thermodynmics, and physics depend on 
# absolute T in Kelvin.

def Lv(T=0):
    #  Latent heat of vaporization as a function of T
    #  Lv(Temp[C]) = 2.501e6 + (Cpv-Cw)*(Temp)  [J/kg]
    #  From Bolton, 1980, MWR, 108, 1046-1053.
    Lvap = L + (Cpv-Cw) * T 
    return Lvap

def es(T,p=1.0e5):
    #   es(T[C],p[Pa]) [Pa] is saturation vapor pressure based on Wexler's formula,
    #   with enhancement factor for moist air rather than water vapor.
    #   The enhancement factor requires a pressure.
    #   T [degrees C], p [Pa] (note the reversed input order), es [Pa]
    #   From A. L. Buck 1981: JAM, 20, 1527-1532.
    #   SPdeS 7 July 2004
    esat = 1e2*6.1121*(1.0007 + 3.46e-8*p)*np.exp((17.502*T)/(240.97 + T))
    return esat

def qs(T,p=1.0e5):
    #  qs(p,T) is saturation specific humidity based on Wexler's formula for es
    #   with enhancement factor (see es.m).
    #   T [degrees C], p [Pa], qs [kg/kg]
    #   From A. L. Buck 1981: JAM, 20, 1527-1532.
    #   SPdeS 7 July 2004
    esat = es(T,p)
    qsat = (Rd/Rv)*esat/(p+(Rd/Rv-1)*esat)
    return qsat

# qsea = lambda T,p: 0.98 * qs(T,p)
def qsea(T,p):
    # seawater specific humidity [kg/kg]
    return 0.98*qs(T,p)

def Tdew(e,p):
    #   Tdew(e,p)
    #   e [Pa] vapor pressure, p pressure [Pa]
    #   Dewpoint Temperature [Celsius] from inversion of es formula of
    #   A. L. Buck 1981: JAM, 20, 1527-1532.
    #   SPdeS 2 August 2010
    a  = 17.502
    b  = 240.97
    k  = np.log( e/6.1121 / (1.0007+3.46e-6*p) ) # depends only on e,p
    Td = b*k / (a-k)
    return Td

def dqsdT(T,p):
    # dqsdT(T[C],p[Pa]) [(kg/kg) / K]
    # saturation specific humidity change with T from Clausius-Clapeyron
    dqsatdT = qs(T,p)*Lv(T) / ( Rv * (T+CtoK)**2 )
    return dqsatdT
    
def dqsdp(Temp,p):
    #  dqsdp(p,Temp) = d(qs)/d(Temp) = -qs(p,Temp)/(p - es(Temp,p))   [(kg/kg) Pa^{-1}]
    #  p[Pa], Temp[degree C]
    dqsatdp = -qs(Temp,p) / (p - es(Temp,p));
    return dqsatdp

def rs(T,p=1.0e5):
    # rs(p,T) is saturation mixing ratio based on Wexler's formula for es
    # with enhancement factor (see es.m).
    # p [Pa], T [degrees C], rs [kg/kg]
    # From A. L. Buck 1981: JAM, 20, 1527-1532.
    # SPdeS 7 July 2004
    esat = es(T,p)
    rsat = (Rd/Rv)*esat/(p-esat)
    return rsat

def Tlcl(Temp,p,qv):
    #   Tl=Tlcl(p[Pa], Temp[K], qv[kg/kg])
    #   Temperature at the LCL
    #   From Bolton, 1980, MWR, 108, 1046-1053.
    ev = p * qv / (Rd/Rv + qv);
    Tl = 2840/(3.5*np.log(Temp) - np.log(0.01*ev) - 4.805) + 55;
    return Tl


def Exner(p,qv=0):
    # Exner function Pi = (p/1.e5)**((Rd/Cp)*(1-0.28*qv))
    Pi = (p/1.e5)**((Rd/Cp)*(1-0.28*qv))
    return Pi

def theta(Temp,p,qv=0):
    #   Potential temperature Temp*(1e5/p)^(Rd/Cp)
    th = Temp / Exner(p,qv)
    return th

def temp(th,p,qv=0):
    # temp(theta[K],p[Pa],qv[kg/kg]=0)
    return th * Exner(p,qv)

def theta_e(Temp,p,qv):
    # theta_e(Temp[K],p[Pa],qv[kg/kg]) from eqn. 43 of Bolton, 1980, MWR, 108, 1046-1053.
    Tl = Tlcl(Temp,p,qv)
    thetae = Temp/Exner(p,qv) * np.exp((3376/Tl - 2.54) * qv * (1 + 0.81*qv))
    return thetae

def theta_es(Temp,p):
    # theta_es(Temp[K],p[Pa],qv[kg/kg]) from eqn. 43 of Bolton, 1980, MWR, 108, 1046-1053.
    return theta_e(Temp,p,qs(Temp-CtoK,p))

def Twet(T,qv,p=1.0e5,niter=5):
    # Tw=Twet(Tdry[C], qv[kg/kg], p[Pa], niter=5)
    # Computes the wet bulb temperature from an isobaric process
    # by using heat internal energy -Cp*dT=L*dq to evaporate
    # and raise qv to qs(Twb).
    #
    # Simon de Szoeke 2016-04-07

    # initialize Tw, qw
    Tw=T
    # problem for temperatures greater than 100 C!
    qw=min(qs(Tw,p),qv)

    # Steps by 1/4 of saturation deficit and enforces -Cp*dT=L*dq by successive approximations
    for iter in range(1,niter+1):
        dq = 0.25 * (qs(p,Tw) - qw)
        if dq < 0:
            break
        dT = -Lv(Tw) / Cp*dq
        qw = qw + dq
        Tw = Tw + dT
        
    return Tw

def theta_w(Temp,p,qv):
    # theta_w(Temp[K],p[Pa],qv[kg/kg]) from eqn 3.8 of Davies-Jones 2008 and 
    # eqn. 43 of Bolton, 1980, MWR, 108, 1046-1053.
    
    # Davies-Jones rational function coefficients
    a0=   7.101574
    a1= -20.68208
    a2=  16.11182
    a3=   2.574631
    a4=  -5.205688
    b1=  -3.552497
    b2=   3.781782
    b3=  -0.6899655
    b4=  -0.5929340
    
    # compute Theta_e from Bolton as first approx for thw
    th = Temp / Exner(p,qv)
    Tl = Tlcl(Temp,p,qv)
    thw = th * np.exp( (3376/Tl - 2.54) *qv * (1 + 0.81*qv) );
    
    # refine thw
    # evaluate Davies-Jones 2008 rational function fit for thw [here in K]
    if thw >= 173.15:
        X = thw / 273.15
        X2 = X  * X
        X3 = X2 * X
        X4 = X2 * X2
        thw = thw - np.exp( (a0+a1*X+a2*X2+a3*X3+a4*X4) / (1+b1*X+b2*X2+b3*X3+b4*X4) );
    
    return thw

def Tv(T,qv):
    # Tv(T[K],qv[kg/kg]) = T * (1 + delta*qv); delta = 0.608
    Tvirt = T * (1 + delta*qv)
    return Tvirt
    
def density(T,p,qv=0):
    # rho = p./(Rd*Tv(T,qv))
    rho = p / (Rd * Tv(T,qv))
    return rho


# lapse rate and sounding methods
    
def gamma(Temp,p):
    # dT/dz = gamma(T[C],p[Pa])
    # moist adiabatic lapse rate [degree C/m]
    gam = ( Lv(Temp) / Cp ) * dqsdT(Temp,p)
    return gam

def Hscale(Temp,qv=0):
    # Hscale(Temp[K],qv[kh/kh]=0) = Rd*Temp/g [m] is pressure scale height
    return Rd * Tv(Temp,qv) / g;
    
def dqsdzs(Temp,p):
    # dqsdzs(p,Temp) = dqs/dz [K/m] on a moist adiabat.
    # Temp [C], p [Pa]
    #
    # dqsatdzs = dqsdzu(Temp,p) / (1 + gamma(Temp,p))
    return dqsdzu(Temp,p) / (1 + gamma(Temp,p))

def dqsdzu(p,Temp):
    #  dqsdzu = dqs/dz [K/m] on a dry adiabat.
    # Temp [C], p [Pa]
    dqsatdzu = ((Rd*(Temp+CtoK)/Cp) * dqsdT(Temp,p) + p * dqsdp(Temp,p)) / Hscale(Temp+CtoK)
    
    return dqsatdzu


def buoyancy_flux(shf, lhf, T=25.0, p=1.0e5, q=0):
    # [bflx, bflx_s]=buoyancy_flux(shf, lhf, T, q, p)  [shf, lhf W/m^2; T °C; q kg/kg; p hPa]
    # Buoyancy flux from sensible and latent heat flux. Optionally also returns the sensible
    # part of the buoyancy flux. bflx = bflx_virt + bflx_s
    #
    # Simon de Szoeke  2018-09-01
    
    virtfac = 1.0 + delta*q
    Tvirt   = (T+CtoK) * virtfac
    rho     = density(Tvirt,p)
    wt      = shf / (rho*Cp)
    wtv     = wt * virtfac + lhf/(rho * Lv(T)) * (T+CtoK)*delta
    bflx    = g * wtv / Tvirt
    bflx_s  = g * wt  / Tvirt
    
    return bflx, bflx_s


# Planck functions
def Planck_la(la,T):
    # B_la=Planck_la(lambda[m],T[K]) [W/sr/m^3]
    # Computes Planck's blackbody spectral radiance as a function of wavelength lambda [m].
    # (c) Simon de Szoeke 2016-04-18
    
    B_la = 2 * hc * clight / ( (la**5) * ( np.exp(hc / (la*kb*T)) - 1 ) ) # W/sr/m^3
    return B_la

def Planck_nu(nu,T):
    # B_nu=Planck_nu(nu[1/s],T[K]) [W/sr/m^3]
    # Computes Planck's blackbody spectral radiance as a function of wavelength lambda [m].
    # (c) Simon de Szoeke 2016-04-18
    B_nu = 2 * hplanck * nu*nu*nu / ( clight*clight * ( np.exp( hplanck*nu / (kb*T) ) - 1 ) ) # W/sr/m^2 s
    return B_nu



# Create __version__ attribute from setup.py information
from pkg_resources import get_distribution, DistributionNotFound
import os.path
from ._version import get_versions

__import__('pkg_resources').declare_namespace(__name__)

try:
    __version__ = get_versions()['version']
except:
    __version__ = '--'
del get_versions



"""
Script to create or adjust necessary config file
"""

import configparser


def read_config(cfg_file):
    """
    read config file
    """
    config = configparser.ConfigParser()
    config.read(cfg_file)
    return config


def rqst_user_input(property, current_value):
    """
    Ask user to modify the property
    given and retuns the modification
    """
    print("Currently: {prop} = {value}".
          format(prop=property,
                 value=current_value))
    update = input("Please modify (leave blank to leave unchanged):")
    if update == '':
        return current_value
    else:
        return update


def adjust_config(cfg_instance):
    """
    Modification of configurations
    by user input
    """
    cfg_sections = cfg_instance.sections()
    for section in cfg_sections:
        for field in cfg_instance[section]:
            update = rqst_user_input(field, cfg_instance[section][field])
            cfg_instance[section][field] = update
    return cfg_instance


def write_config(cfg_instance, cfg_file):
    """
    Write updated configuration back to config
    file
    """
    print("Writing updated config file to {}".format(cfg_file))
    with open(cfg_file, 'w') as configfile:
        cfg_instance.write(configfile)


def update_config(cfg_file, cfg_file_new_location):
    """
    Update config file or create a personal copy
    in the user space
    """
    current_config = read_config(cfg_file)
    updated_config = adjust_config(current_config)
    write_config(updated_config, cfg_file_new_location)


if __name__ == '__main__':
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Script to convert ASCII files of radiosonde type RS41

Original version by: Johannes Kiliani/Lukas Frank
"""

# insert some subroutines if possible
import time
import datetime
import calendar
import os.path
import sys
import glob
import subprocess as sp
import configparser
from configparser import ExtendedInterpolation
import argparse
import logging
import numpy as np
from pathlib import Path
import netCDF4
from netCDF4 import Dataset, default_fillvals, num2date

sys.path.append(os.path.dirname(os.path.abspath(__file__)))
from config import cfg_creator as configupdater

try:
    import eurec4a_snd
    __version__ = eurec4a_snd.__version__
except ModuleNotFoundError:
    print('Not found')
    __version__ = 'see git_version'

# ====================================================
# General MPI-BCO settings:
# ====================================================


def load_configuration(configuration_file=None):
    """
    Loads the configuration file PATH.ini.
    1. If provided load configuration_file
    2. Attempt to load from home directory
    3. Attempt to load from relative path inside BCO-git structure

    Args:
        configuration_file: optional: complete path to the configuration file.

    Returns:
        instance of ConfigParser class with extended interpolation.
    """
    dir_path = os.path.dirname(os.path.realpath(__file__))
    ini_path = "/".join(dir_path.split("/")[:-1]) + "/eurec4a_snd/config/meta_information.ini"
    if not isinstance(configuration_file, str):
        possible_file_in_userdir = Path("~/meta_information.ini").expanduser()
        if os.path.isfile(possible_file_in_userdir):
            configuration_file = possible_file_in_userdir
        elif os.path.isfile(ini_path):
            configuration_file = ini_path
        if configuration_file is None or not os.path.isfile(configuration_file):
            raise FileNotFoundError(
                "No Configuration File 'meta_information.ini' found. Please create one"
                " in your home directory "
                "or provide the path via the argument parsing -c.")
        else:
            logging.info("Using configuration file: %s" % configuration_file)

    conf = configparser.ConfigParser(interpolation=ExtendedInterpolation())
    conf.read(configuration_file)
    return conf


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--configfile', metavar="meta_information.ini", help='Provide a meta_information.ini configuration file. \n'
                                                                       'If not provided it will be searched for at:\n'
                                                                       '1. ~/meta_information.ini\n'
                                                                       '2. ../../../meta_information.ini', required=False, default=None)
    parser.add_argument("-i", "--inputfile", metavar="INPUT_FILE",
                        help="Single rs41 data file (txt) or file format\n"
                             "including wildcards", default=None, required=False)

    parser.add_argument("-p", "--inputpath", metavar='/some/example/path/',
                        help="Path to the folder containing radiosonde txt files",
                        default=None,
                        required=False)

    parser.add_argument("-o", "--outputfolder", metavar="/some/example/path/",
                        help="Output folder for converted files (netCDF)",
                        default=None,
                        required=False)

    parser.add_argument('-z', '--compression', metavar="COMPRESSION_LEVEL",
                        help="Set the Level of compression for the output (1-9)",
                        required=False, default=4)

    parser.add_argument('-v', '--verbose', metavar="DEBUG",
                        help='Set the level of verbosity [DEBUG, INFO,'
                        ' WARNING, ERROR]',
                        required=False, default="INFO")

    parser.add_argument('-d', '--date', metavar="YYYYMMDD", help='Provide the desired date to be processed. '
                                                                 'Fomat: YYYYMMDD', required=False, default=None)
    parsed_args = vars(parser.parse_args())

    if (parsed_args["date"] is None) and (parsed_args["inputfile"] is None):
        parser.error(
            "either --date or --inputfile needs to be set. (--date not yet implemented)")

    return parsed_args


def setup_logging(verbose):
    assert verbose in ["DEBUG", "INFO", "WARNING", "ERROR"]
    logging.basicConfig(
        level=logging.getLevelName(verbose),
        format="%(levelname)s - %(name)s - %(funcName)s - %(message)s",
        handlers=[
            logging.FileHandler("{}.log".format(__file__)),
            logging.StreamHandler()
        ])

def main():
    # Set up global configuration of BCO-MPI-GIT:
    args = get_args()
    setup_logging(args['verbose'])

    try:
        config = load_configuration(args["configfile"])
    except FileNotFoundError:
        logging.info('No configuration file could be found and will now'
                     ' be created with your help')
        configupdater.update_config(os.path.abspath(os.path.dirname(__file__)) +
            '/config/meta_information_template.ini',
            Path('~/meta_information.ini').expanduser())
        sys.exit("Config file has been created at {0}. Please restart script with the option -c {0}".format(Path('~/meta_information.ini').expanduser()))
        if args["outputfolder"] is None and (args["inputpath"] is None and args["inputpath"] is None):
            sys.exit("No config file found! Outputfolder and Inputpath"
                     " or Inputfile need to be provided!")
        else:
            logging.warning("The file meta_information.ini could not be found and"
                            " no metainformation will be added to the output! "
                            " This is not recommended!")
            pass
    else:
        if args["inputpath"] is None:
            args["inputpath"] = config["FILES"]["INPUT_DAT"]
        if args["outputfolder"] is None:
            args["outputfolder"] = config["FILES"]["OUTPUT_DAT2NC"]

    try:
        git_module_version = sp.check_output(
            ["git", "describe", "--always"]).strip()
    except:
        logging.info('No git-version could be found. Please consider'
                     'pulling the git repository.')
        git_module_version = "--"

    compression_level = int(args["compression"])

    time_in = time.time()

    # Creating file list according to given arguments
    if args['inputfile'] is None:
        filelist = glob.glob(args['inputpath'] + '*.txt')
    else:
        filelist = glob.glob(args['inputfile'])
    filelist = sorted(filelist)

    logging.info('Files to process {}'.format(filelist))
    for ifile in range(0, len(filelist)):
        logging.info('Reading file number {}'.format(ifile))
        linecount = 0
        with open(filelist[ifile], 'rb') as fa:
            for line in fa:  # loop over all lines within ASCII file
                logging.info(line.decode('ISO-8859-1'))
                linecount += 1  # Counts all lines within header
                string = line.split()
                if linecount == 1:
                    soundid = string[-1]
                if linecount == 2:
                    datest = string[-1][0:4] + string[-1][5:7] + string[-1][8:10]
                    timest = string[-1][11:13] + \
                        string[-1][14:16] + string[-1][17:19]
                if linecount == 3:
                    serial = string[-1]
                if linecount == 4:
                    sondetype = string[-1]
                if linecount == 10:
                    break

        utctime = calendar.timegm(datetime.datetime(int(datest[0:4]), int(datest[4:6]), int(datest[6:8]),
                                                    int(timest[0:2]), int(timest[2:4]), int(timest[4:6])).timetuple())
        data = np.genfromtxt(
            filelist[ifile], skip_header=9, invalid_raise=False, encoding="ISO-8859-1")

        num_rows = data.shape[0]  # length of raw array
        logging.info('Number of rows: {}'.format(num_rows))

        # after all needed header information is read, the reduced data field is masked for
        # NaN values and an output file produced afterward:

        tindex = np.ma.masked_invalid(data[:, 1])
        vvert_m = np.ma.masked_invalid(data[:, 9])
        alti_m = np.ma.masked_invalid(data[:, 2])
        pres_m = np.ma.masked_invalid(data[:, 3])
        temp_m = np.ma.masked_invalid(data[:, 4])
        rh_m = np.ma.masked_invalid(data[:, 5])
        # mixing ratio is not including in input file
        # but remains for compatibility with airport
        # sounding files
        mix_m = np.ma.masked_invalid(data[:, 5])
        dewp_m = np.ma.masked_invalid(data[:, 6])
        vhori_m = np.ma.masked_invalid(data[:, 8])
        vdir_m = np.ma.masked_invalid(data[:, 7])
        lat_m = np.ma.masked_invalid(data[:, 11])
        long_m = np.ma.masked_invalid(data[:, 12])

        # calculate water vapor mixing ratio from temperature and
        # relative humidity using magnus formula
        vapor_pressure = (rh_m/100.) * (611.2 * np.exp((17.62*(temp_m))/(243.12 + temp_m)))
        mix_m = 1000.*((0.622*vapor_pressure)/(100.*pres_m - vapor_pressure))

        rh_m = np.ma.masked_outside(rh_m, 0., 100.)
        pres_m = np.ma.masked_less(pres_m, 5.)
        vhori_m = np.ma.masked_greater(vhori_m, 100.)
        mix_m = np.ma.masked_greater(mix_m, -1., )

        # Find temporal resolution
        # using most common time difference
        _, indices = np.unique(np.diff(tindex), return_inverse=True)
        time_resolution = (np.diff(tindex)[np.argmax(np.bincount(indices))])

        # Create outputfile with time information from file
        sounding_date = num2date(
            np.int(utctime), 'seconds since 1970-01-01 00:00:00 UTC')
        YYYYMM = sounding_date.strftime('%Y%m')
        YYYYMMDDHHMM = sounding_date.strftime('%Y%m%d%H%M')

        outpath = args['outputfolder'] + YYYYMM + '/'
        if not os.path.exists(outpath):
            success = sp.call(["mkdir", "-p", outpath])

        outfile = outpath + \
            "{platform}_Sounding{direction}_{location}_{date}.nc".\
            format(platform=config['PLATFORM']['platform_name_short'],
                   location=config['PLATFORM']['platform_location'].
                                replace(' ', '').
                                replace(',', '').
                                replace(';', ''),
                   direction='AscentProfile',
                   date=sounding_date.strftime('%Y%m%d_%H%M'))

        # Creation of output NetCDF file
        fo = Dataset(outfile, 'w', format='NETCDF4')

        # assign NetCDF file attributes from strings read from ASCII file header
        fo.title = 'Sounding data containing temperature, pressure, humidity,' \
                   ' latitude, longitude, wind direction, wind speed, and time'
        # Platform information
        fo.platform_name = '{long} ({short})'.format(
            long=config['PLATFORM']['platform_name_long'],
            short=config['PLATFORM']['platform_name_short'])
        fo.location = config['PLATFORM']['platform_location']
        fo.surface_altitude = config['PLATFORM']['platform_altitude']

        # Instrument metadata
        fo.instrument = config['INSTRUMENT']['instrument_description']
        fo.number_of_Probe = serial
        fo.radiosonde_type = sondetype

        # Information about launch
        fo.date_YYYYMMDD = datest
        fo.time_of_launch_HHmmss = timest
        fo.launch_unixtime = utctime
        fo.latitude_of_launch_location = '{0:5.2f} deg N'.format(np.round(lat_m[0], 2))
        fo.longitude_of_launch_location = '{0:6.2f} deg E'.format(np.round(long_m[0], 2))

        # Information about output
        fo.resolution = "{:g} sec".format(time_resolution)
        fo.source = filelist[ifile]
        fo.git_version = git_module_version
        fo.created_with = '{file} with its last modifications on {time}'.\
            format(time=time.ctime(os.path.getmtime(os.path.realpath(__file__))),
                   file=os.path.basename(__file__))
        fo.created_on = str(time.ctime(time.time()))
        fo.contact_person = '{name} ({mail})'.format(
            name=config['OUTPUT']['contact_person_name'],
            mail=config['OUTPUT']['contact_person_email'])
        fo.institution = config['OUTPUT']['institution']
        fo.converted_by = '{name} ({mail})'.format(
            name=config['OUTPUT']['executive_person_name'],
            mail=config['OUTPUT']['executive_person_email'])
        fo.python_version = "{} (with numpy:{}, netCDF4:{}, eurec4a_snd:{})".\
            format(sys.version, np.__version__, netCDF4.__version__, __version__)
        fo.Conventions = 'CF-1.7'
        fo.featureType = "trajectory"

        # Define Dimension (record length) from ASCII record counter
        fo.createDimension('levels', num_rows)
        prof_dim = fo.createDimension('trajectory', None)
        str_dim = fo.createDimension('str_dim', 1000)
        fillval = default_fillvals['f4']

        # Creation of NetCDF Variables, including description and unit
        nc_prof = fo.createVariable(
            'trajectory', 'S1', ('trajectory', 'str_dim'), fill_value='', zlib=True)
        nc_prof.cf_role = "trajectory_id"
        nc_prof.long_name = 'trajectory identifier'
        nc_prof.description = 'unique string describing the trajectories origin'

        nc_launchtime = fo.createVariable('launch_time', 'f8', ('trajectory'), zlib=True)
        nc_launchtime.long_name = "time at which the sonde has been launched"
        nc_launchtime.units = 'seconds since 1970-01-01 00:00:00 UTC'
        nc_launchtime.calendar = 'gregorian'
        nc_launchtime.standard_name = 'time'

        nc_tindex = fo.createVariable(
            'flight_time', 'f4', ('trajectory', 'levels'), fill_value=fillval, zlib=True)
        nc_tindex.long_name = 'time passed since launch'
        nc_tindex.standard_name = 'time'
        nc_tindex.units = 'seconds since {launch}'.format(
            launch=sounding_date.strftime('%Y-%m-%d %H:%M%S UTC'))
        nc_tindex.axis = 'T'
        nc_tindex.calendar = "gregorian"
        nc_vvert = fo.createVariable(
            'ascentRate', 'f4', ('trajectory', 'levels'), fill_value=fillval, zlib=True)
        nc_vvert.long_name = 'ascent/descent rate of balloon or other measuring device'
        nc_vvert.description = 'ascent rate is positive/ descent rate is negative'
        nc_vvert.units = 'm/s'
        nc_vvert.coordinates = "flight_time longitude latitude pressure"
        nc_alti = fo.createVariable(
            'altitude', 'f4', ('trajectory', 'levels'), fill_value=fillval, zlib=True)
        nc_alti.standard_name = 'altitude'
        nc_alti.units = 'm'
        nc_alti.coordinates = "flight_time longitude latitude pressure"
        nc_pres = fo.createVariable(
            'pressure', 'f4', ('trajectory', 'levels'), fill_value=fillval, zlib=True)
        nc_pres.standard_name = 'air_pressure'
        nc_pres.units = 'hPa'
        nc_pres.axis = 'Z'
        nc_pres.positive = 'down'
        nc_temp = fo.createVariable(
            'temperature', 'f4', ('trajectory', 'levels'), fill_value=fillval, zlib=True)
        nc_temp.standard_name = 'air_temperature'
        nc_temp.units = 'degrees_Celsius'
        nc_temp.coordinates = "flight_time longitude latitude pressure"
        nc_rh = fo.createVariable(
            'humidity', 'f4', ('trajectory', 'levels'), fill_value=fillval, zlib=True)
        nc_rh.standard_name = 'relative_humidity'
        nc_rh.units = '%'
        nc_rh.coordinates = "flight_time longitude latitude pressure"
        nc_dewp = fo.createVariable(
            'dewPoint', 'f4', ('trajectory', 'levels'), fill_value=fillval, zlib=True)
        nc_dewp.standard_name = 'dew_point_temperature'
        nc_dewp.units = 'degrees_Celsius'
        nc_dewp.coordinates = "flight_time longitude latitude pressure"
        nc_mix = fo.createVariable(
            'mixingRatio', 'f4', ('trajectory', 'levels'), fill_value=fillval, zlib=True)
        nc_mix.long_name = 'water vapor mixing ratio'
        nc_mix.standard_name = 'humidity_mixing_ratio'
        nc_mix.units = 'g/kg'
        nc_mix.coordinates = "flight_time longitude latitude pressure"
        nc_vhori = fo.createVariable(
            'windSpeed', 'f4', ('trajectory', 'levels'), fill_value=fillval, zlib=True)
        nc_vhori.standard_name = 'wind_speed'
        nc_vhori.units = 'm/s'
        nc_vhori.coordinates = "flight_time longitude latitude pressure"
        nc_vdir = fo.createVariable(
            'windDirection', 'f4', ('trajectory', 'levels'), fill_value=fillval, zlib=True)
        nc_vdir.standard_name = 'wind_from_direction'
        nc_vdir.units = 'degrees'
        nc_vdir.coordinates = "flight_time longitude latitude pressure"
        nc_lat = fo.createVariable(
            'latitude', 'f4', ('trajectory', 'levels'), fill_value=fillval, zlib=True)
        nc_lat.long_name = 'latitude'
        nc_lat.standard_name = 'latitude'
        nc_lat.units = 'degrees_north'
        nc_lat.axis = 'Y'
        nc_long = fo.createVariable(
            'longitude', 'f4', ('trajectory', 'levels'), fill_value=fillval, zlib=True)
        nc_long.long_name = 'longitude'
        nc_long.standard_name = 'longitude'
        nc_long.units = 'degrees_east'
        nc_long.axis = 'X'

        trajectory_name = '{platform}__{lat}_{lon}__{launchtime}'.\
                          format(platform=config['PLATFORM']['platform_name_short'],
                                 lat=lat_m[0],
                                 lon=long_m[0],
                                 launchtime=str(YYYYMMDDHHMM))
        trajectory_name_parts = []
        for char in trajectory_name:
            trajectory_name_parts.extend(char)

        nc_prof[0, 0:len(trajectory_name_parts)] = trajectory_name_parts
        nc_launchtime[0] = utctime

        nc_tindex[0, :] = tindex[:]
        nc_vvert[0, :] = vvert_m[:]
        nc_alti[0, :] = alti_m[:]
        nc_pres[0, :] = pres_m[:]
        nc_temp[0, :] = temp_m[:]
        nc_rh[0, :] = rh_m[:]
        nc_dewp[0, :] = dewp_m[:]
        nc_mix[0, :] = mix_m[:]
        nc_vhori[0, :] = vhori_m[:]
        nc_vdir[0, :] = vdir_m[:]
        nc_lat[0, :] = lat_m[:]
        nc_long[0, :] = long_m[:]

        fo.close()
        logging.info('DONE: {input} converted to {output}'.format(
            input=filelist[ifile],
            output=outfile))

    time_out = time.time()
    logging.debug('System time: {}s'.format(time_out - time_in))


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Script to convert ASCII files of radiosonde type RS41

Original version by: Johannes Kiliani/Lukas Frank
"""

# insert some subroutines if possible
import time
import platform
from pathlib import Path, PureWindowsPath
import shutil
import os.path
import sys
import subprocess as sp
import configparser
from configparser import ExtendedInterpolation
import argparse
import logging
import numpy as np
import netCDF4
from netCDF4 import Dataset, default_fillvals, num2date, date2num

sys.path.append(os.path.dirname(os.path.abspath(__file__)))
from config import cfg_creator as configupdater
from _helpers import *
import _thermo as thermo

# ====================================================
# General MPI-BCO settings:
# ====================================================


def load_configuration(configuration_file=None):
    """
    Loads the configuration file PATH.ini.
    1. If provided load configuration_file
    2. Attempt to load from home directory
    3. Attempt to load from relative path inside BCO-git structure

    Args:
        configuration_file: optional: complete path to the configuration file.

    Returns:
        instance of ConfigParser class with extended interpolation.
    """
    dir_path = os.path.dirname(os.path.realpath(__file__))
    ini_path = "/".join(dir_path.split("/")[:-1]) + "/eurec4a_snd/config/meta_information.ini"
    if not isinstance(configuration_file, str):
        possible_file_in_userdir = Path("~/meta_information.ini").expanduser()
        if os.path.isfile(possible_file_in_userdir):
            configuration_file = possible_file_in_userdir
        elif os.path.isfile(ini_path):
            configuration_file = ini_path
        if configuration_file is None or not os.path.isfile(configuration_file):
            raise FileNotFoundError(
                "No Configuration File 'meta_information.ini' found. Please create one"
                " in your home directory "
                "or provide the path via the argument parsing -c.")
        else:
            logging.info("Using configuration file: %s" % configuration_file)

    conf = configparser.ConfigParser(interpolation=ExtendedInterpolation())
    conf.read(configuration_file)
    return conf


def get_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('-c', '--configfile', metavar="meta_information.ini", help='Provide a meta_information.ini configuration file. \n'
                                                                       'If not provided it will be searched for at:\n'
                                                                       '1. ~/meta_information.ini\n'
                                                                       '2. ../../../meta_information.ini', required=False, default=None)
    parser.add_argument("-i", "--inputfile", metavar="INPUT_FILE",
                        help="Single sonde file (bufr) or file format\n"
                             "including wildcards",
                        default=None,
                        required=False,
                        type=unixpath)

    parser.add_argument("-p", "--inputpath", metavar='/some/example/path',
                        help="Path to the folder containing sonde bufr files.",
                        default=None,
                        required=False,
                        type=unixpath)

    parser.add_argument("-o", "--outputfolder", metavar="/some/example/path/",
                        help="Output folder for converted files (netCDF). You can\n"
                             " although not recommended also define an output file\n"
                             "(format). However, please share only those with the\n"
                             " the default filename.\n"
                             " The following formats can be used:\n"
                             "\t {platform}\t platform name\n"
                             "\t {location}\t platform location\n"
                             "\t {direction}\t sounding direction\n"
                             "\t {date}\t\t date of sounding release\n"
                             "\t\t or self defined date format with\n"
                             "\t\t %%Y%%m%%d %%H%%M and so on\n"
                             "\t\t and others to format the output folder dynamically.",
                        default=None,
                        required=False)

    parser.add_argument('-v', '--verbose', metavar="DEBUG",
                        help='Set the level of verbosity [DEBUG, INFO,'
                        ' WARNING, ERROR]',
                        required=False, default="INFO")

    parsed_args = vars(parser.parse_args())

    if (parsed_args["inputpath"] is not None) and (parsed_args["inputfile"] is not None):
        parser.error(
            "either --inputpath or --inputfile should be used. Use --inputpath"
            "for several files and --inputfile for single ones")

    if (parsed_args["inputpath"] is None) and (parsed_args["inputfile"] is None):
        parser.error(
            "either --inputpath or --inputfile must be defined. Use --inputpath"
            "for several files and --inputfile for single ones.")

    return parsed_args


def unixpath(path_in):
    """
    Convert windows path to unix path syntax
    depending on the used OS
    """
    if platform.system() == 'Windows':
        path_out = Path(PureWindowsPath(path_in))
    else:
        path_out = Path(path_in)
    return path_out


def setup_logging(verbose):
    assert verbose in ["DEBUG", "INFO", "WARNING", "ERROR"]
    logging.basicConfig(
        level=logging.getLevelName(verbose),
        format="%(levelname)s - %(name)s - %(funcName)s - %(message)s",
        handlers=[
            logging.FileHandler("{}.log".format(__file__)),
            logging.StreamHandler()
        ])


def main():
    # Set up global configuration of BCO-MPI-GIT:
    args = get_args()

    setup_logging(args['verbose'])

    logging.debug('Gathering version information')
    try:
        import eurec4a_snd
        __version__ = eurec4a_snd.__version__
        package_version_set = True
    except (ModuleNotFoundError, AttributeError):
        logging.debug('No eurec4a_snd package version found')
        __version__ = 'see git_version'
        package_version_set = False

    try:
        git_module_version = sp.check_output(
            ["git", "describe", "--always", "--dirty"], stderr=sp.STDOUT).strip()
        git_version_set = True
    except sp.CalledProcessError:
        logging.debug('No git-version could be found.')
        git_module_version = "--"
        git_version_set = False

    if (not git_version_set and not package_version_set):
        logging.warning('No version of the converter could be found!'
                        ' Please consider the installation via conda'
                        ' or if this is not working clone the git re'
                        'pository')

    logging.info('Version of script: {} (conda package), {} (git version)'.format(__version__, git_module_version))

    logging.debug('Search for configuration file and load in case present')
    try:
        config = load_configuration(args["configfile"])
    except FileNotFoundError:
        logging.info('No configuration file could be found and will now'
                     ' be created with your help')
        configupdater.update_config(os.path.abspath(os.path.dirname(__file__)) +
            '/config/meta_information_template.ini',
            Path('~/meta_information.ini').expanduser())
        sys.exit("Config file has been created at {0}. Please restart script with the option -c {0}".format(Path('~/meta_information.ini').expanduser()))
        if args["outputfolder"] is None and (args["inputpath"] is None and args["inputpath"] is None):
            sys.exit("No config file found! Outputfolder and Inputpath"
                     " or Inputfile need to be provided!")
        else:
            logging.warning("The file meta_information.ini could not be found and"
                            " no metainformation will be added to the output! "
                            " This is not recommended!")
            pass
    else:
        if args["inputpath"] is None:
            args["inputpath"] = config["FILES"]["INPUT_DAT"]
        if args["outputfolder"] is None:
            args["outputfolder"] = config["FILES"]["OUTPUT_DAT2NC"]

    time_in = time.time()
    date_unit = "seconds since 1970-01-01 00:00:00 UTC"

    logging.debug('Create filelist')
    if args['inputfile'] is None:
        filelist = args['inputpath'].glob('*.bfr')
    else:
        filelist = [args['inputfile']]
    filelist = sorted(filelist)

    logging.info('Files to process {}'.format([file.name for file in filelist]))
    for ifile, bufr_file in enumerate(filelist):
        logging.info('Reading file number {}'.format(ifile))

        json_file = convert_bufr_to_json(bufr_file, logging)
        json_flat, keys = read_json(json_file)
        if args['verbose'] != 'DEBUG':
            shutil.rmtree(os.path.dirname(json_file))

        sounding = convert_json_to_arrays(json_flat, keys)
        sounding = replace_missing_data(sounding)
        sounding = convert_list_to_array(sounding)

        sounding.latitude = calculate_coordinates(
            sounding.station_lat,
            sounding.displacement_lat)
        sounding.longitude = calculate_coordinates(
            sounding.station_lon,
            sounding.displacement_lon)

        sounding = bufr_specific_handling(sounding)

        try:
            serial = sounding.meta_data['sonde_serial_number']
        except KeyError:
            serial = '--'
        try:
            sondetype = sounding.meta_data['radiosondeType']
        except KeyError:
            sondetype = '--'
        try:
            sondefreq = sounding.meta_data['sonde_frequency']
        except KeyError:
            sondefreq = '--'

        logging.debug('Get sounding direction (ascending/descending)')
        sounding.direction = get_sounding_direction(sounding.meta_data['bufr_msg'])
        if sounding.direction == 1:
            # Upward
            direction_str = 'Ascent'
        elif sounding.direction == -1:
            # Downward
            direction_str = 'Descent'

        sounding = expected_unit_check(sounding)

        # after all needed header information is read, the reduced data field
        # is masked for NaN values and an output file produced afterward:
        logging.debug('Mask invalid sounding data')
        sounding.time = np.ma.masked_invalid(sounding.time)
        sounding.gpm = np.ma.masked_invalid(sounding.gpm)
        sounding.pressure = np.ma.masked_invalid(sounding.pressure)
        sounding.temperature = np.ma.masked_invalid(sounding.temperature)
        sounding.dewpoint = np.ma.masked_invalid(sounding.dewpoint)
        sounding.windspeed = np.ma.masked_invalid(sounding.windspeed)
        sounding.winddirection = np.ma.masked_invalid(sounding.winddirection)
        sounding.latitude = np.ma.masked_invalid(sounding.latitude)
        sounding.longitude = np.ma.masked_invalid(sounding.longitude)

        # Calculate additional variables
        e_sat = thermo.es(sounding.temperature, sounding.pressure*100)
        e = thermo.es(sounding.dewpoint, sounding.pressure*100)
        sounding.relativehumidity = (e/e_sat)*100
        sounding.mixingratio = (thermo.Rd/thermo.Rv)*e/(sounding.pressure*100-e)*1000

        # Ascent rate
        sounding = calc_ascentrate(sounding)

        # Sort sounding by flight time
        sounding = sort_sounding_by_time(sounding)

        # Remove 1000hPa reduced gpm
        sounding = exclude_1000hPa_gpm(sounding)

        # Find temporal resolution
        time_resolution = calc_temporal_resolution(sounding)

        # Create outputfile with time information from file
        sounding_date = sounding.sounding_start_time
        YYYYMMDDHHMM = sounding.sounding_start_time.strftime('%Y%m%d%H%M')

        outpath_fmt = unixpath(args['outputfolder'])
        outpath = Path(sounding.sounding_start_time.strftime(outpath_fmt.as_posix()))

        if outpath.suffix == '.nc':
            outfile = Path(outpath.as_posix().format(platform=config['PLATFORM']['platform_name_short'],
                                                     location=config['PLATFORM']['platform_location'].
                                                               replace(' ', '').
                                                               replace(',', '').
                                                               replace(';', ''),
                                                     direction='{}Profile'.format(direction_str),
                                                     date=sounding_date.strftime('%Y%m%d_%H%M')))
        else:
            outfile = Path(os.path.join(outpath, \
                "{platform}_Sounding{direction}_{location}_{date}.nc".\
                format(platform=config['PLATFORM']['platform_name_short'],
                       location=config['PLATFORM']['platform_location'].
                                 replace(' ', '').
                                 replace(',', '').
                                 replace(';', ''),
                       direction='{}Profile'.format(direction_str),
                       date=sounding_date.strftime('%Y%m%d_%H%M'))))

        if not outfile.parent.exists():
            os.makedirs(outfile.parent)

        # Creation of output NetCDF file
        fo = Dataset(outfile, 'w', format='NETCDF4')

        # assign NetCDF file attributes from meta data
        fo.title = 'Sounding data containing temperature, pressure, humidity,'\
                   ' latitude, longitude, wind direction, wind speed, and time'
        # Platform information
        fo.platform_name = '{long} ({short})'.format(
            long=config['PLATFORM']['platform_name_long'],
            short=config['PLATFORM']['platform_name_short'])
        fo.location = config['PLATFORM']['platform_location']
        fo.surface_altitude = config['PLATFORM']['platform_altitude']

        # Instrument metadata
        fo.instrument = config['INSTRUMENT']['instrument_description']
        fo.number_of_Probe = serial
        fo.sonde_type = sondetype
        fo.sonde_frequency = sondefreq

        # Information about launch
        fo.date_YYYYMMDD = sounding_date.strftime('%Y%m%d')
        fo.time_of_launch_HHmmss = sounding_date.strftime('%H%M%S')
        fo.launch_unixtime = date2num(sounding.sounding_start_time, date_unit)
        fo.latitude_of_launch_location = '{0:5.2f} deg N'.\
            format(sounding.station_lat)
        fo.longitude_of_launch_location = '{0:6.2f} deg E'.\
            format(sounding.station_lon)

        # Information about output
        fo.resolution = "{:g} sec".format(time_resolution)
        fo.source = Path(PureWindowsPath(bufr_file)).absolute().as_posix()
        fo.git_version = git_module_version
        fo.created_with = '{file} with its last modifications on {time}'.\
            format(time=time.ctime(os.path.getmtime(os.path.realpath(__file__))),
                   file=os.path.basename(__file__))
        fo.created_on = str(time.ctime(time.time()))
        fo.contact_person = '{name} ({mail})'.format(
            name=config['OUTPUT']['contact_person_name'],
            mail=config['OUTPUT']['contact_person_email'])
        fo.institution = config['OUTPUT']['institution']
        fo.converted_by = '{name} ({mail})'.format(
            name=config['OUTPUT']['executive_person_name'],
            mail=config['OUTPUT']['executive_person_email'])
        fo.python_version = "{} (with numpy:{}, netCDF4:{}, eurec4a_snd:{})".\
            format(sys.version, np.__version__, netCDF4.__version__,
                   __version__)
        fo.Conventions = 'CF-1.7'
        fo.featureType = "trajectory"

        # Define Dimension (record length) from ASCII record counter
        fo.createDimension('levels', len(sounding.pressure))
        fo.createDimension('sounding', None)
        fo.createDimension('str_dim', 1000)
        fillval = default_fillvals['f4']

        # Creation of NetCDF Variables, including description and unit
        nc_prof = fo.createVariable(
            'sounding', 'S1', ('sounding', 'str_dim'),
            fill_value='',
            zlib=True)
        nc_prof.cf_role = "sounding_id"
        nc_prof.long_name = 'sounding identifier'
        nc_prof.description = 'unique string describing the soundings origin'

        nc_launchtime = fo.createVariable('launch_time', 'f8', ('sounding'),
            zlib=True)
        nc_launchtime.long_name = "time at which the sonde has been launched"
        nc_launchtime.units = 'seconds since 1970-01-01 00:00:00 UTC'
        nc_launchtime.calendar = 'gregorian'
        nc_launchtime.standard_name = 'time'

        nc_tindex = fo.createVariable(
            'flight_time', 'f4', ('sounding', 'levels'),
            fill_value=fillval,
            zlib=True)
        nc_tindex.long_name = 'time passed since launch'
        nc_tindex.standard_name = 'time'
        nc_tindex.units = 'seconds since {launch}'.format(
            launch=sounding_date.strftime('%Y-%m-%d %H:%M:%S UTC'))
        nc_tindex.axis = 'T'
        nc_tindex.calendar = "gregorian"
        nc_vvert = fo.createVariable(
            'ascentRate', 'f4', ('sounding', 'levels'),
            fill_value=fillval,
            zlib=True)
        nc_vvert.long_name = 'ascent/descent rate of balloon or other measuring device'
        nc_vvert.description = 'ascent rate is positive/ descent rate is negative'
        nc_vvert.units = 'm/s'
        nc_vvert.coordinates = "flight_time longitude latitude pressure"
        nc_alti = fo.createVariable(
            'altitude', 'f4', ('sounding', 'levels'),
            fill_value=fillval,
            zlib=True)
        nc_alti.standard_name = 'altitude'
        nc_alti.units = 'm'
        nc_alti.coordinates = "flight_time longitude latitude pressure"
        nc_pres = fo.createVariable(
            'pressure', 'f4', ('sounding', 'levels'),
            fill_value=fillval,
            zlib=True)
        nc_pres.standard_name = 'air_pressure'
        nc_pres.units = 'hPa'
        nc_pres.axis = 'Z'
        nc_pres.positive = 'down'
        nc_temp = fo.createVariable(
            'temperature', 'f4', ('sounding', 'levels'),
            fill_value=fillval,
            zlib=True)
        nc_temp.standard_name = 'air_temperature'
        nc_temp.units = 'degrees_Celsius'
        nc_temp.coordinates = "flight_time longitude latitude pressure"
        nc_rh = fo.createVariable(
            'humidity', 'f4', ('sounding', 'levels'),
            fill_value=fillval,
            zlib=True)
        nc_rh.standard_name = 'relative_humidity'
        nc_rh.units = '%'
        nc_rh.coordinates = "flight_time longitude latitude pressure"
        nc_dewp = fo.createVariable(
            'dewPoint', 'f4', ('sounding', 'levels'),
            fill_value=fillval,
            zlib=True)
        nc_dewp.standard_name = 'dew_point_temperature'
        nc_dewp.units = 'degrees_Celsius'
        nc_dewp.coordinates = "flight_time longitude latitude pressure"
        nc_mix = fo.createVariable(
            'mixingRatio', 'f4', ('sounding', 'levels'),
            fill_value=fillval,
            zlib=True)
        nc_mix.long_name = 'water vapor mixing ratio'
        nc_mix.standard_name = 'humidity_mixing_ratio'
        nc_mix.units = 'g/kg'
        nc_mix.coordinates = "flight_time longitude latitude pressure"
        nc_vhori = fo.createVariable(
            'windSpeed', 'f4', ('sounding', 'levels'),
            fill_value=fillval,
            zlib=True)
        nc_vhori.standard_name = 'wind_speed'
        nc_vhori.units = 'm/s'
        nc_vhori.coordinates = "flight_time longitude latitude pressure"
        nc_vdir = fo.createVariable(
            'windDirection', 'f4', ('sounding', 'levels'),
            fill_value=fillval,
            zlib=True)
        nc_vdir.standard_name = 'wind_from_direction'
        nc_vdir.units = 'degrees'
        nc_vdir.coordinates = "flight_time longitude latitude pressure"
        nc_lat = fo.createVariable(
            'latitude', 'f4', ('sounding', 'levels'),
            fill_value=fillval,
            zlib=True)
        nc_lat.long_name = 'latitude'
        nc_lat.standard_name = 'latitude'
        nc_lat.units = 'degrees_north'
        nc_lat.axis = 'Y'
        nc_long = fo.createVariable(
            'longitude', 'f4', ('sounding', 'levels'),
            fill_value=fillval,
            zlib=True)
        nc_long.long_name = 'longitude'
        nc_long.standard_name = 'longitude'
        nc_long.units = 'degrees_east'
        nc_long.axis = 'X'

        sounding_name = '{platform}__{lat:5.2f}_{lon:5.2f}__{launchtime}'.\
                        format(platform=config['PLATFORM']['platform_name_short'],
                               lat=sounding.station_lat,
                               lon=sounding.station_lon,
                               launchtime=str(YYYYMMDDHHMM))
        sounding_name_parts = []
        for char in sounding_name:
            sounding_name_parts.extend(char)

        nc_prof[0, 0:len(sounding_name_parts)] = sounding_name_parts
        nc_launchtime[0] = date2num(sounding.sounding_start_time, date_unit)

        nc_tindex[0, :] = sounding.time
        nc_vvert[0, :] = sounding.ascentrate
        nc_alti[0, :] = sounding.gpm
        nc_pres[0, :] = sounding.pressure
        nc_temp[0, :] = sounding.temperature
        nc_rh[0, :] = sounding.relativehumidity
        nc_dewp[0, :] = sounding.dewpoint
        nc_mix[0, :] = sounding.mixingratio
        nc_vhori[0, :] = sounding.windspeed
        nc_vdir[0, :] = sounding.winddirection
        nc_lat[0, :] = sounding.latitude
        nc_long[0, :] = sounding.longitude

        fo.close()
        logging.info('DONE: {input} converted to {output}'.format(
            input=filelist[ifile],
            output=outfile))

    time_out = time.time()
    logging.debug('System time: {}s'.format(time_out - time_in))


if __name__ == "__main__":
    main()



import setuptools
import versioneer

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="eurec4a_snd",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="eurec4a folks",
    author_email="",
    description="Common EUREC4A sounding standard",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/observingClouds/eurec4a_snd",
    packages=setuptools.find_packages(),
    setup_requires=['setuptools-git-version'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=['pillow>=6.0.0', 'matplotlib>=3.1.0', 'basemap>=1.2.0',
                      'numpy>=1.15.0', 'netCDF4>=1.4.0', 'metpy>=0.10.0'],
    entry_points={'console_scripts':
                    ['sounding_converter=eurec4a_snd.L1_bufr:main',
                     'sounding_visualize=eurec4a_snd.make_quicklooks_rs41:main',
                     'sounding_skewT=eurec4a_snd.visualize.make_skewT_metpy:main']},
    package_data={"eurec4a_snd": ["examples/data/*", "config/meta_information_template.ini"]}
)



"""
Helper functions for bufr import
"""
import tempfile
import os
import json
import datetime as dt
import numpy as np


class UnitChangedError(Exception):
    pass


class UnexpectedUnit(Exception):
    pass


def convert_bufr_to_json(bufr_fn, logger=None):
    """
    Convert bufr file to json with ecCodes
    software
    """
    tmp_folder = tempfile.mkdtemp()
    tmp_output_json_fn = os.path.join(tmp_folder, 'tmp_bufr.json')
    r = os.system("bufr_dump -j s {} > {}".format(bufr_fn, tmp_output_json_fn))
    if logger is None:
        print("Converted {} to {}".format(bufr_fn, tmp_output_json_fn))
    else:
        logger.debug("Converted {} to {}".format(bufr_fn, tmp_output_json_fn))
    return tmp_output_json_fn


def flatten_json(y):
    """
    Flatten structured json array
    
    Input
    -----
    y : list or dict

    Return
    ------
    list : list
        list containing dicts for each key value pair
    """
    global r
    out = {}
    r = 0

    def flatten(x, name='', number=0):
        """helper function"""
        global r
        if type(x) is dict:
            for a in x:
                flatten(x[a], a)
        elif type(x) is list:
            i = 0
            for a in x:
                number = flatten(a)
                i += 1
                r += 1
        else:
            out['{:07g}_{}'.format(r, name)] = x

    flatten(y)
    return out


def read_json(json_fn):
    """
    Read and flatten json
    """
    with open(json_fn) as file:
        struct_json = json.load(file)

    bfr_json_flat = flatten_json(struct_json)
    keys = bfr_json_flat.keys()
    key_keys = []
    for key in keys:
        if 'key' in key:
            key_keys.append(key[:-4])

    return bfr_json_flat, key_keys


def convert_json_to_arrays(json_flat, key_keys):
    """
    Convert json data to array
    """
    class Sounding:
        """
        Class containing sounding data
        """
        def __init__(self):
            self.station_lat = None
            self.station_lon = None
            self.sounding_start_time = None
            self.time = []
            self.time_unit = None
            self.pressure = []
            self.pressure_unit = None
            self.temperature = []
            self.temperature_unit = None
            self.dewpoint = []
            self.dewpoint_unit = None
            self.windspeed = []
            self.windspeed_unit = None
            self.winddirection = []
            self.winddirection_unit = None
            self.gpm = []
            self.gpm_unit = None
            self.displacement_lat = []
            self.displacement_lat_unit = None
            self.displacement_lon = []
            self.displacement_lon_unit = None
            self.meta_data = {}

    def _ensure_measurement_integrity(self):
        """
        Test integrity of each measurement unit

        Measurements of the sonde in the bufr file
        contain usually:
            - time since launch
            - pressure
            - gpm
            - location displacement
            - temperature
            - dewpoint
            - wind direction
            - wind speed

        This is a complete unit. However, there are,
        dependining on the bufr format additional
        measurements, which might not consist of
        a complete measurement set. This is for ex.
        the case for the entry which contains the
        "absoluteWindShearIn1KmLayerBelow"

        Here the completeness of the measurement
        is checked and corrected otherwise by
        adding nan values to the not measured
        values.
        """
        if len(self.time) > len(self.temperature):
            self.temperature.append(np.nan)
        if len(self.time) > len(self.gpm):
            self.gpm.append(np.nan)
        if len(self.time) > len(self.dewpoint):
            self.dewpoint.append(np.nan)
        if len(self.time) > len(self.pressure):
            self.pressure.append(np.nan)
        if len(self.time) > len(self.windspeed):
            self.windspeed.append(np.nan)
        if len(self.time) > len(self.winddirection):
            self.winddirection.append(np.nan)
        if len(self.time) > len(self.displacement_lat):
            self.displacement_lat.append(np.nan)
        return

    s = Sounding()

    for key_key in key_keys:
        if json_flat[key_key+'_key'] == 'latitude':
            s.station_lat = json_flat[key_key+'_value']
        elif json_flat[key_key+'_key'] == 'longitude':
            s.station_lon = json_flat[key_key+'_value']
        elif json_flat[key_key+'_key'] == 'pressure':
            s.pressure.append(json_flat[key_key+'_value'])
            if s.pressure_unit is None:
                s.pressure_unit = json_flat[key_key+'_units']
            # Unit consistency test
            elif s.pressure_unit != json_flat[key_key+'_units']:
                raise UnitChangedError('{} and {} are not same unit'.format(s.pressure_unit,
                                                                            json_flat[key_key+'_units']))
        elif json_flat[key_key+'_key'] == 'windSpeed':
            s.windspeed.append(json_flat[key_key+'_value'])
            if s.windspeed_unit is None:
                s.windspeed_unit = json_flat[key_key+'_units']
            # Unit consistency test
            elif s.windspeed_unit != json_flat[key_key+'_units']:
                raise UnitChangedError('{} and {} are not same unit'.format(s.windspeed_unit,
                                                                            json_flat[key_key+'_units']))
        elif json_flat[key_key+'_key'] == 'windDirection':
            s.winddirection.append(json_flat[key_key+'_value'])
            if s.winddirection_unit is None:
                s.winddirection_unit = json_flat[key_key+'_units']
            # Unit consistency test
            elif s.winddirection_unit != json_flat[key_key+'_units']:
                raise UnitChangedError('{} and {} are not same unit'.format(s.winddirection_unit,
                                                                            json_flat[key_key+'_units']))
        elif json_flat[key_key+'_key'] == 'nonCoordinateGeopotentialHeight':
            s.gpm.append(json_flat[key_key+'_value'])
            if s.gpm_unit is None:
                s.gpm_unit = json_flat[key_key+'_units']
            # Unit consistency test
            elif s.gpm_unit != json_flat[key_key+'_units']:
                raise UnitChangedError('{} and {} are not same unit'.format(s.gpm_unit,
                                                                            json_flat[key_key+'_units']))
        elif json_flat[key_key+'_key'] == 'airTemperature':
            s.temperature.append(json_flat[key_key+'_value'])
            if s.temperature_unit is None:
                s.temperature_unit = json_flat[key_key+'_units']
            # Unit consistency test
            elif s.temperature_unit != json_flat[key_key+'_units']:
                raise UnitChangedError('{} and {} are not same unit'.format(s.temperature_unit,
                                                                            json_flat[key_key+'_units']))
        elif json_flat[key_key+'_key'] == 'dewpointTemperature':
            s.dewpoint.append(json_flat[key_key+'_value'])
            if s.dewpoint_unit is None:
                s.dewpoint_unit = json_flat[key_key+'_units']
            # Unit consistency test
            elif s.dewpoint_unit != json_flat[key_key+'_units']:
                raise UnitChangedError('{} and {} are not same unit'.format(s.dewpoint_unit,
                                                                            json_flat[key_key+'_units']))
        elif json_flat[key_key+'_key'] == 'latitudeDisplacement':
            s.displacement_lat.append(json_flat[key_key+'_value'])
            if s.displacement_lat_unit is None:
                s.displacement_lat_unit = json_flat[key_key+'_units']
            # Unit consistency test
            elif s.displacement_lat_unit != json_flat[key_key+'_units']:
                raise UnitChangedError('{} and {} are not same unit'.format(s.displacement_lat_unit,
                                                                            json_flat[key_key+'_units']))
        elif json_flat[key_key+'_key'] == 'longitudeDisplacement':
            s.displacement_lon.append(json_flat[key_key+'_value'])
            if s.displacement_lon_unit is None:
                s.displacement_lon_unit = json_flat[key_key+'_units']
            # Unit consistency test
            elif s.displacement_lon_unit != json_flat[key_key+'_units']:
                raise UnitChangedError('{} and {} are not same unit'.format(s.displacement_lon_unit,
                                                                            json_flat[key_key+'_units']))
        elif json_flat[key_key+'_key'] == 'timePeriod':
            # Check conistency of data
            _ensure_measurement_integrity(s)
            s.time.append(json_flat[key_key+'_value'])

            if s.time_unit is None:
                s.time_unit = json_flat[key_key+'_units']
            # Unit consistency test
            elif s.time_unit != json_flat[key_key+'_units']:
                raise UnitChangedError('{} and {} are not same unit'.format(s.time_unit,
                                                                            json_flat[key_key+'_units']))
        elif json_flat[key_key+'_key'] == 'year':
            year = json_flat[key_key+'_value']
        elif json_flat[key_key+'_key'] == 'month':
            month = json_flat[key_key+'_value']
        elif json_flat[key_key+'_key'] == 'day':
            day = json_flat[key_key+'_value']
        elif json_flat[key_key+'_key'] == 'hour':
            hour = json_flat[key_key+'_value']
        elif json_flat[key_key+'_key'] == 'minute':
            minute = json_flat[key_key+'_value']
        elif json_flat[key_key+'_key'] == 'second':
            second = json_flat[key_key+'_value']

        # Meta data
        elif json_flat[key_key+'_key'] == 'radiosondeSerialNumber':
            s.meta_data['sonde_serial_number'] = json_flat[key_key+'_value']
        elif json_flat[key_key+'_key'] == 'softwareVersionNumber':
            s.meta_data['softwareVersionNumber'] = json_flat[key_key+'_value']
        elif json_flat[key_key+'_key'] == 'radiosondeType':
            s.meta_data['radiosondeType'] = json_flat[key_key+'_value']
        elif json_flat[key_key+'_key'] == 'unexpandedDescriptors':
            try:
                s.meta_data['bufr_msg'] = json_flat[key_key+'_value']
            except KeyError:
                # Error probably caused, because there are several
                # unexpandedDescriptors
                # which seems to be only the case for dropsondes?!
                s.meta_data['bufr_msg'] = 309053
        elif json_flat[key_key+'_key'] == 'radiosondeOperatingFrequency':
            s.meta_data['sonde_frequency'] = str(json_flat[key_key+'_value']) + json_flat[key_key+'_units']

    _ensure_measurement_integrity(s)

    s.sounding_start_time = dt.datetime(year,
                                        month,
                                        day,
                                        hour,
                                        minute,
                                        second)

    return s


def replace_missing_data(sounding):
    """
    Removing Nones from sounding measurements
    """
    def replace_none(entry):
        """
        Replace None with NaN
        """
        if entry == None:
            return np.nan
        else:
            return entry

    variables = ['displacement_lat', 'displacement_lon', 'pressure', 'windspeed',
                 'winddirection', 'temperature', 'dewpoint', 'gpm', 'time']

    for var in variables:
        sounding.__dict__[var] = list(map(replace_none, sounding.__dict__[var]))

    return sounding


def convert_list_to_array(sounding):
    """
    Convert datatype of sounding
    """
    variables = ['displacement_lat', 'displacement_lon', 'pressure', 'windspeed',
                 'winddirection', 'temperature', 'dewpoint', 'gpm', 'time']

    for var in variables:
        sounding.__dict__[var] = np.array(sounding.__dict__[var])

    return sounding


def calculate_coordinates(origin, offset):
    """
    Calculate positon of measurement

    Input
    -----
    origin : float
        latitude or longitude of launch position
    offset : float
        latitudinal or longitudinal displacement
        since launch from origin

    Return
    ------
    float : position of measurement
    """
    return origin + offset


def calc_ascentrate(sounding):
    """
    Calculate the ascent rate

    Input
    -----
    sounding : obj
        sounding class containing gpm
        and flight time

    Return
    ------
    soundning : obj
        sounding including the ascent rate
    """
    ascent_rate = np.diff(sounding.gpm)/(np.diff(sounding.time))
    ascent_rate = np.ma.concatenate(([0], ascent_rate))  # 0 at first measurement
    sounding.ascentrate = ascent_rate

    return sounding


def calc_temporal_resolution(sounding):
    """
    Calculate temporal resolution of sounding

    Returns the most common temporal resolution
    by calculating the temporal differences
    and returning the most common difference.

    Input
    -----
    sounding : obj
        sounding class containing flight time
        information

    Return
    ------
    temporal_resolution : float
        temporal resolution
    """
    time_differences = np.abs(np.diff(np.ma.compressed(sounding.time)))
    time_differences_counts = np.bincount(time_differences.astype(int))
    most_common_diff = time_differences[np.argmax(time_differences_counts)]
    temporal_resolution = most_common_diff
    return temporal_resolution


def bufr_specific_handling(sounding):
    """
    Apply bufr message specific functions

    Depending on the BUFR format, that data
    has to be prepared differently

    BUFR309053 (dropsonde)
    BUFR309056 (radiosonde descent)
    BUFR309057 (radiosonde ascent)
    - Remove last entries of time, latitude, longitude because those
        belong to the 'absoluteWindShearIn1KmLayerAbove'/
        'absoluteWindShearIn1KmLayerBelow' entries

    """
    variables = ['latitude', 'longitude', 'pressure', 'windspeed',
                 'winddirection', 'temperature', 'dewpoint', 'gpm', 'time']

    if sounding.meta_data['bufr_msg'] == 309053:
        # Nothing to do so far
        pass
    elif sounding.meta_data['bufr_msg'] == 309056:
        # Nothing to do so far
        pass
    elif sounding.meta_data['bufr_msg'] == 309057:
        # Nothing to do so far
        pass
    return sounding


def get_sounding_direction(bufr_msg):
    """
    Get direction of sounding

    1: upward
    -1: downward
    """

    if str(bufr_msg) == '309053':
        return -1
    elif str(bufr_msg) == '309056':
        return -1
    elif str(bufr_msg) == '309057':
        return 1
    else:
        raise NotImplementedError('The bufr message format {} is not implemented'.format(bufr_msg))


def kelvin_to_celsius(kelvin):
    """
    Convert Kelvin to Celsius
    """
    return kelvin - 273.15


def pascal_to_hectoPascal(pascal):
    """
    Convert Pa to hPa
    """
    return pascal/100.


converter_dict = {'K-->C': kelvin_to_celsius,
                  'K-->degC': kelvin_to_celsius,
                  'Pa-->hPa': pascal_to_hectoPascal
                  }


def calc_relative_humidity(sounding):
    """
    Calculate relative humidity
    """
    relative_humidity = 100*(np.exp((17.625*sounding.dewpoint)/
        (243.04+sounding.dewpoint))/np.exp((17.625*sounding.temperature)/
        (243.04+sounding.temperature)))
    return relative_humidity


def calc_vapor_pressure(sounding):
    """
    Calculate water vapor pressure
    """
    vapor_pressure = (sounding.relativehumidity/100.) * (611.2 * np.exp((17.62 * (sounding.temperature))/(243.12 + sounding.temperature)))
    return vapor_pressure


def calc_wv_mixing_ratio(sounding, vapor_pressure):
    """
    Calculate water vapor mixing ratio
    """
    wv_mix_ratio = 1000.*((0.622*vapor_pressure)/(100.*sounding.pressure - vapor_pressure))
    return wv_mix_ratio


def expected_unit_check(sounding):
    """
    Check if units are as expected
    and try to convert accordingly
    """

    variables = ['displacement_lat', 'displacement_lon', 'pressure', 'windspeed',
                 'winddirection', 'temperature', 'dewpoint', 'gpm', 'time']

    expected_bufr_units = ['deg', 'deg', 'Pa', 'm/s', 'deg', 'K', 'K', 'gpm', 's']
    expected_output_units = ['deg', 'deg', 'hPa', 'm/s', 'deg', 'degC', 'degC', 'gpm', 's']

    for v, var in enumerate(variables):
        if (sounding.__dict__[var+'_unit'] != expected_output_units[v]):
            # Convert data to expected unit
            ## Find converter function
            try:
                func = converter_dict['-->'.join([sounding.__dict__[var+'_unit'], expected_output_units[v]])]
            except KeyError:
                raise UnexpectedUnit('Unit {} was expected, but got {}. Conversion was not successful'.format(expected_bufr_units[v],
                                                                           sounding.__dict__[var+'_unit']))
            else:
                sounding.__dict__[var] = func(sounding.__dict__[var])
                sounding.__dict__[var+'_unit'] = expected_output_units[v]
        else:
            pass

    return sounding


def nan_argsort(array, direction=1):
    """
    Sorting with handeling nan values
    depending on the sounding direction

    Input
    -----
    array : array-like
        data to sort
    direction : int
        integer (-1, 1) to indicate on which end
        of the sorted array nan values should be
        saved.

    Result
    ------
    indices : array
        Indices that would sort the input array
    """
    tmp = array.copy().astype('float')
    tmp[np.isnan(array)] = -np.inf*direction
    return np.argsort(tmp)


def sort_sounding_by_time(sounding):
    """
    Sort sounding by altitude
    """
    sorter = nan_argsort(sounding.time, sounding.direction)
    sounding.time = sounding.time[sorter]
    sounding.ascentrate = sounding.ascentrate[sorter]
    sounding.gpm = sounding.gpm[sorter]
    sounding.pressure = sounding.pressure[sorter]
    sounding.temperature = sounding.temperature[sorter]
    sounding.relativehumidity = sounding.relativehumidity[sorter]
    sounding.dewpoint = sounding.dewpoint[sorter]
    sounding.mixingratio = sounding.mixingratio[sorter]
    sounding.windspeed = sounding.windspeed[sorter]
    sounding.winddirection = sounding.winddirection[sorter]
    sounding.latitude = sounding.latitude[sorter]
    sounding.longitude = sounding.longitude[sorter]

    return sounding


def exclude_1000hPa_gpm(sounding):
    """
    BUFR files include values calculated for 1000 hPa
    even when the sounding starts at an higher
    elevation.

    These values are those, where time contain
    a missing value.

    This function returns the sounding
    without the missing data in the time
    dimension.
    """
    nan_mask = ~np.isnan(sounding.time)
    sounding.time = sounding.time[nan_mask]
    sounding.ascentrate = sounding.ascentrate[nan_mask]
    sounding.gpm = sounding.gpm[nan_mask]
    sounding.pressure = sounding.pressure[nan_mask]
    sounding.temperature = sounding.temperature[nan_mask]
    sounding.relativehumidity = sounding.relativehumidity[nan_mask]
    sounding.dewpoint = sounding.dewpoint[nan_mask]
    sounding.mixingratio = sounding.mixingratio[nan_mask]
    sounding.windspeed = sounding.windspeed[nan_mask]
    sounding.winddirection = sounding.winddirection[nan_mask]
    sounding.latitude = sounding.latitude[nan_mask]
    sounding.longitude = sounding.longitude[nan_mask]
    return sounding
